#!/usr/bin/env python
# coding=utf-8


import http.client
import json
import paramiko
import os
from influxdb import InfluxDBClient

class checkKvm:

    def __init__(self):
        self.Host = []
        self.centerData = {}
        self.whiteList = []
        self.sshKeyFile = paramiko.RSAKey.from_private_key_file("/root/.ssh/id_rsa")
        self.alert_text = ""
        self.flag = 0
        self.dbClient = ""
        self.num = 0


    def getCenterData(self):
        try:
            url = "simba.gyyx.cn"
            uri = "/get_kvm_machine"
            conn = http.client.HTTPConnection(url)
            conn.request(method="GET", url="http://" + url + uri)
            response = conn.getresponse()
            ress = response.read()
            resultsDict = json.loads(json.loads(ress)['data'])
            for i in resultsDict:
                inData = 0
                pm_list = []
                if i["p_gysn"] in self.whiteList: continue
                if i["p_ip"] not in self.Host:
                    self.Host.append(i["p_ip"])
                # 创建1个存放宿主机和虚拟机对应关系的字典，宿主机ip作为键。
                if i["p_ip"] not in self.centerData:
                    self.centerData[i["p_ip"]] = []
                for pmHost in self.centerData[i["p_ip"]]:
                    for vmgysn in pmHost.values():
                        if vmgysn[1] == i['gysn']:
                            inData = 1
                            break
                if inData == 0:
                    gy_pm_dict = {i['gysn']: [i['p_gysn'], i['ip'], i['manager']]}
                    self.centerData[i["p_ip"]].append(gy_pm_dict)
        except Exception as e:
            print(e)

    def kvmOperate(self):
        try:
            for ip in self.Host:
                sshClient = self.connectHost(ip)
                # 如果连接宿主机报错则写入数据库：宿主机连接异常
                if sshClient == 1:
                    self.writeDb(ip, [0,0], [0,0], [0,0],0)
                    continue

                # 检测是否有虚拟机没有设置开机自启动并写入数据库
                self.hostAutoStatus(sshClient,ip)

                # 检测宿主机各项资源情况
                diskResList = self.hostDisk(sshClient)
                cpuResList = self.hostCpu(sshClient)
                memResList = self.hostMem(sshClient)
                self.writeDb(ip, diskResList, cpuResList, memResList)


                # 检测已关机状态虚拟机: 占着数据目录、状态为shut off，写入数据库
                downHost = self.hostStatus(sshClient)
                if downHost != 1:
                    result = str(downHost).replace('\\n','')
                    self.writeOffHost(ip, result)

            # print("关机虚拟机数量：",self.num)

            # if self.alert_text:
            #     cmd_sa = """curl 'https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=c1de31ff-5cd7-4daf-a53f-9e097aa9a2fc' -H 'Content-Type: application/json' -d '{"msgtype": "text","text": {"content": "A502机房KVM虚拟机开机自启置项检测报警:\n%s","mentioned_list":["@all"]}}'""" % (self.alert_text)
            #     os.system(cmd_sa)
            if self.flag == 0:
                cmd = """curl -i -XPOST 'http://influxdb.ops.gyyx.cn:8086/write?db=sa' --data-binary 'vm_autostart,agysn=none,bvm_name=none,cmanager=none,tags=0 values=0i';"""
                os.system(cmd)
        except Exception as e:
            print(e)

    def hostAutoStatus(self,sshclient,ip):
        try:
            stdin_running, stdout_running, stderr_running = sshclient.exec_command("virsh list  | sed -n '/running/p' | awk '{print $2}'")
            stdin_autostart, stdout_autostart, stderr_autostart = sshclient.exec_command(" ls  /etc/libvirt/qemu/autostart/ | sort -k1.5n ")
            runningHost = stdout_running.readlines()
            autostartHost = str(stdout_autostart.readlines()).replace('.xml', '').split()
            # auto_vm_list:用来存放已配置开机自启的机器，若running状态下的机器没设置开机自启，则报警。
            autoHostList = []
            for auto_vm in autostartHost:
                auto_vm_gysn = str(auto_vm).strip("[],'u").replace(' ', '')[:-2]
                autoHostList.append(auto_vm_gysn)
            for run_vm in runningHost:
                run_vm_strip = run_vm.strip()
                if run_vm_strip not in autoHostList:
                    # 推送消息到grafana监控
                    for asset in self.centerData[ip]:
                        for vmgysn in asset.keys():
                            vmgysn =vmgysn
                        if vmgysn == run_vm_strip:
                            self.flag = 1
                            suzhuji_gysn = asset[run_vm_strip][0]
                            vm_gysn = run_vm_strip
                            vm_manager = asset[run_vm_strip][2]
                            if not vm_manager:
                                vm_manager = "中心应用无对应负责人"
                            cmd = """curl -i -XPOST 'http://influxdb.ops.gyyx.cn:8086/write?db=sa' --data-binary 'vm_autostart,agysn=%s,bvm_name=%s,cmanager=%s,tags=1 values=1i';""" % (suzhuji_gysn, vm_gysn, vm_manager)
                            os.system(cmd)
                            self.alert_text += "宿主机：" + suzhuji_gysn + ", " + "虚拟机：" + vm_gysn + ", " + "负责人：" + vm_manager + "\n"
        except Exception as e:
            print(e)


    def hostDisk(self,sshclient):
        try:
            stdinDiskSpace, stdoutDiskSpace, stderrDiskSpace = sshclient.exec_command("""df -ThP /data1  |grep -E "^/dev/" |awk '{print $(NF-2)}'""")
            stdinDiskUse, stdoutDiskUse, stderrDiskUse = sshclient.exec_command("""df -ThP /data1  |grep -E "^/dev/" |awk '{print $(NF-1)}'""")
            DiskSpace = str(stdoutDiskSpace.readlines()[0]).strip('\n')
            DiskUse = str(stdoutDiskUse.readlines()[0]).strip('\n')
            return DiskSpace,DiskUse
        except Exception as e:
            print(e)

    def hostCpu(self,sshclient):
        try:
            stdinCpuUse, stdoutCpuUse, stderrCpuUse = sshclient.exec_command("""top -b -n 1 | grep -w "Cpu(s):" |awk '{print $2}'|awk -F '%' '{print $1}'""")
            stdinCpuCore, stdoutCpuCore, stderrCpuCore = sshclient.exec_command("""lscpu  |grep "^CPU(s)" |awk '{print $2}'""")
            CpuUse = str(stdoutCpuUse.readlines()[0]).strip('\n')
            CpuCore = int(stdoutCpuCore.readlines()[0])
            return CpuCore,CpuUse+"%"
        except Exception as e:
            print(e)

    def hostMem(self,sshclient):
        try:
            stdinMemTotal, stdoutMemTotal, stderrMemTotal = sshclient.exec_command( """grep -w  'MemTotal:' /proc/meminfo | awk '{print $2}'""")
            stdinMemFree, stdoutMemFree, stderrMemFree = sshclient.exec_command( """grep -w  'MemFree:' /proc/meminfo | awk '{print $2}'""")
            stdinMemBuffers, stdoutMemBuffers, stderrMemBuffers = sshclient.exec_command( """grep  -w 'Buffers:' /proc/meminfo | awk '{print $2}'""")
            stdinMemCached, stdoutMemCached, stderrMemCached = sshclient.exec_command( """grep  -w 'Cached:' /proc/meminfo | awk '{print $2}'""")
            MemTotal = int(stdoutMemTotal.readlines()[0])
            MemFree = int(stdoutMemFree.readlines()[0])
            MemBuffers = int(stdoutMemBuffers.readlines()[0])
            MemCached = int(stdoutMemCached.readlines()[0])
            used = int(MemFree + MemBuffers + MemCached)
            return str(round(MemTotal/1024/1024))+"G",str(round(used/1024/1024))+"G"
        except Exception as e:
            print(e)

    def initInfluxdb(self):
        try:
            client = InfluxDBClient("10.14.3.127", "8086", "root", "", 'sa')
            dbcmd1 = "drop measurement vm_autostart;"
            dbcmd2 = "drop measurement kvm_resources;"
            dbcmd3 = "drop measurement kvmOff;"
            client.query(dbcmd1)
            client.query(dbcmd2)
            client.query(dbcmd3)
        except Exception as ex:
            print("influxdb ERROR: %s") %ex
            client = InfluxDBClient("10.14.3.127", "8086", "root", "", 'sa')
        finally:
            self.dbClient = client


    def writeDb(self,ip,diskResList,cpuResList,memResList,flag=1):
        try:
            diskFreeSpase = diskResList[0]
            diskUse = diskResList[1]
            cpuCore = cpuResList[0]
            cpuUse = cpuResList[1]
            memTotla = memResList[0]
            memUse = memResList[1]
            for asset in self.centerData[ip]:
                for vmgysn in asset.values():
                    p_gysn = vmgysn[0]
            json_data = [
                {
                    "measurement": "kvm_resources",
                    "tags":{
                        "B_hostIp":ip,
                        "A_hostGysn":p_gysn,
                        "C_diskFreeSpase":diskFreeSpase,
                        "D_diskUse":diskUse,
                        "E_cpuCore":cpuCore,
                        "F_cpuUse":cpuUse,
                        "G_memTotla":memTotla,
                        "H_memUse":memUse
                    },
                    "fields": {
                        "flag": flag
                    }
                }
            ]
            self.dbClient.write_points(json_data)
        except Exception as e:
            print(e)
        finally:
            self.dbClient.close()

    def connectHost(self,ip):
        try:
            # 如果宿主机可以正常连接，返回ssh对象，否则返回1
            sshClient = paramiko.SSHClient()
            sshClient.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            sshClient.connect(ip, 63008, username="root", pkey=self.sshKeyFile, timeout=3)
            return sshClient
        except:
            return 1

    def hostStatus(self,sshclient):
        try:
            # 获取状态为shut off的虚拟机
            stdinHostStat, stdoutHostStat, stderrHostStat = sshclient.exec_command("""virsh list --all |grep -w 'shut off' |awk '{print $2}'""")

            # 获取存在数据目录但是没有启动的虚拟机
            stdinRunHost, stdoutRunHost, stderrRunHost = sshclient.exec_command("""virsh list  | sed -n '/running/p' | awk '{print $2}'""")
            stdinDirHost, stdoutDirHost, stderrDirHost = sshclient.exec_command("""ls -l /data1/domains/ |grep '^d' |awk '{print $NF}'""")
            HostStat = stdoutHostStat.readlines()
            RunHost = stdoutRunHost.readlines()
            DirHost = stdoutDirHost.readlines()
            diffRes = set(DirHost).difference(set(RunHost))
            result = set(list(diffRes) + HostStat)
            self.num += len(result)
            if result:
                return result
            else:
                return 1
        except Exception as e:
            print(e)

    def writeOffHost(self,ip,data):
        try:
            for asset in self.centerData[ip]:
                for vmgysn in asset.values():
                    p_gysn = vmgysn[0]

            json_data = [
                {
                    "measurement": "kvmOff",
                    "tags": {
                        "B_hostIp": ip,
                        "A_hostGysn": p_gysn,
                        "C_OffHost": data,
                    },
                    "fields": {
                        "flag": 1
                    }
                }
            ]
            self.dbClient.write_points(json_data)
        except Exception as e:
            print(e)

if __name__ == '__main__':
    ob = checkKvm()
    # 初始化数据库
    ob.initInfluxdb()
    # 获取中心数据
    ob.getCenterData()
    # 主逻辑: 循环宿主机，进行虚拟机自启动、关机和宿主机资源相关检测
    ob.kvmOperate()




